/*     
 * ip_control.c
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: ip_control.c,v $
 * Revision 1.27.1.5.1.1  1995/12/01 03:03:20  ho
 * Restored IP FAKING option; add DELIVER_ALL for forwarding.
 *
 * Revision 1.27.1.1  1994/10/11  19:36:22  menze
 * Cleaner separation of local outgoing address and incoming address,
 * reporting the latter as result in GETMYHOST
 *
 */

#include "xkernel.h"
#include "ip.h"
#include "route.h"
#include "ip_i.h"
#ifdef IPSEC
#include "ipsec.h"
#include "crypt.h"
#endif /* IPSEC */


#define IPHOSTLEN	sizeof(IPhost)


/*
 * ip_controlsessn
 */
int
ipControlSessn(s, opcode, buf, len)
    XObj s;
    int opcode;
    char *buf;
    int len;
{
    SState        *sstate;
    PState        *pstate;
    IPheader      *hdr;
    
    xAssert(xIsSession(s));
    
    sstate = (SState *)s->state;
    pstate = (PState *)s->myprotl->state;
    
    hdr = &(sstate->hdr);
    switch (opcode) {
	
      case GETMYHOST :
	checkLen(len, IPHOSTLEN);
	*(IPhost *)buf = sstate->rcvAddr;
	return IPHOSTLEN;
	
      case GETPEERHOST :
	checkLen(len, IPHOSTLEN);
	*(IPhost *)buf = sstate->hdr.dest;  
	return IPHOSTLEN;
	
      case GETMYHOSTCOUNT:
      case GETPEERHOSTCOUNT:
	checkLen(len, sizeof(int));
	*(int *)buf = 1;
	return sizeof(int);

      case GETMYPROTO :
      case GETPEERPROTO :
	checkLen(len, sizeof(long));
	*(long *)buf = sstate->hdr.prot;
	return sizeof(long);
	
      case GETMAXPACKET :
	checkLen(len, sizeof(int));
	*(int *)buf = IPMAXPACKET;
	return sizeof(int);
	
      case GETOPTPACKET :
	checkLen(len, sizeof(int));
	*(int *)buf = sstate->mtu - IPHLEN;
#ifdef IPSEC
	if (sstate->ipsec) {
	  int size = *(int *)buf;
	  int delta = xControl(sstate->ipsec, GETOPTPACKET, 0, 0);
	  *(int *)buf = size - (delta>0?delta:0);
	}
#endif /* IPSEC */
	return sizeof(int);
	
      case IP_REDIRECT:
	return ipControlProtl(s->myprotl, opcode, buf, len);

      case IP_CHANGEROUTE:
	return ipControlProtl(s->myprotl, opcode, buf, len);

      case IP_SETSOURCEADDR:
      {
	checkLen(len, sizeof(IPhost));
	sstate->hdr.source = *(IPhost *) buf;
	return 0;
      }
	
      case GETPARTICIPANTS:
	{
#ifdef IPSEC
	    Part	p[SPI_PART+1];
#else
	    Part	p[LOCAL_PART+1];
#endif
	    int		retLen;

	    retLen = xControl(xGetDown(s, 0), opcode, buf, len);
	    if ( retLen > 0 && partExtLen(buf) > 0 && partExtLen(buf) <= SPI_PART ) {
		partInternalize(p, buf);
		/* 
		 * We may have rewritten the remote participant to be
		 * a gateway, so we'll replace it with the ultimate
		 * destination address.  The local participant (if
		 * it's there) should be OK.
		 */
		if ( partPop(p[0]) == 0 ) {
		    return -1;
		}
		partPush(p[0], &sstate->hdr.dest, sizeof(IPhost));
#ifdef IPSEC
		if (sstate->ipsec) {
		  int retLen2 = len;

		  return (partExternalize(p, buf, &len) == XK_FAILURE) ?
		    -1: xControl(sstate->ipsec, opcode, buf, retLen2);
		}
#endif /* IPSEC */
		return (partExternalize(p, buf, &len) == XK_FAILURE) ?
		  		-1 : len;
	    } else {
		return -1;
	    }
	}

      case IP_GETPSEUDOHDR:
	{
	    IPpseudoHdr	*phdr = (IPpseudoHdr *)buf;

	    checkLen(len, sizeof(IPpseudoHdr));
	    phdr->src = sstate->hdr.source;
	    phdr->dst = sstate->hdr.dest;
	    phdr->zero = 0;
	    phdr->len = 0;
	    phdr->prot = sstate->hdr.prot;
	    return sizeof(IPpseudoHdr);
	}

#ifdef IPSEC
      case IP_PSEUDOHDR:   /* Turn on flag saying to fixup length on IPpseudoheader. */
	if (sstate->ipsec)
	  return xControl(sstate->ipsec, opcode, buf, len);
	else return 0;

      case CRYPT_ISPRIVATE:
	if (sstate->ipsec)
	  return xControl(sstate->ipsec, opcode, buf, len);
	else goto ondown;

      case IPSEC_GETPEERHOST:
	if (sstate->ipsec) {
	  checkLen(len, IPHOSTLEN);
	  *(IPhost *)buf = sstate->hdr.dest;  
	  return IPHOSTLEN; }
	else return -1;

 ondown:  /* must always be before default */

#endif /* IPSEC */

      default : 
	xTrace0(ipp,TR_ERRORS,"Unhandled opcode -- forwarding");
	return xControl(xGetDown(s, 0), opcode, buf, len);
    }
}



/*
 * ip_controlprotl
 */
int
ipControlProtl(self, opcode, buf, len)
    XObj self;
    int opcode;
    char *buf;
    int len;
{
    PState	*pstate;
    IPhost 	net, mask, gw, dest;
    
    xAssert(xIsProtocol(self));
    pstate = (PState *) self->state;
    
    switch (opcode) {
	
      case IP_REDIRECT :
	{
	    checkLen(len, 2*IPHOSTLEN);
	    net = *(IPhost *)buf;
	    netMaskFind(&mask, &net);
	    gw = *(IPhost *)(buf + IPHOSTLEN);
	    xTrace3(ipp, 4, "IP_REDIRECT : net = %s, mask = %s, gw = %s",
		    ipHostStr(&net), ipHostStr(&mask), ipHostStr(&gw));
	    /*
	     * find which interface reaches the gateway
	     */
	    rt_add(pstate, &net, &mask, &gw, -1, RTDEFAULTTTL);
	    return 0;
	}

      case IP_CHANGEROUTE :
        {
	    IProuteInfo *ri;
            checkLen(len, sizeof(IProuteInfo));
	    ri = (IProuteInfo *)buf;
            xTrace5(ipp, TR_EVENTS, "IP_CHANGEROUTE : "
		    "net = %s, mask = %s, gw = %s metric = %d, ttl = %d",
                    ipHostStr(&ri->net), ipHostStr(&ri->mask),
		    ipHostStr(&ri->gw), ri->metric, ri->ttl);
            /*
             * find which interface reaches the gateway
             */
            rt_add(pstate, &ri->net, &ri->mask, &ri->gw, ri->metric, ri->ttl);
            return 0;
        }


      case GETMAXPACKET :
	checkLen(len, sizeof(int));
	*(int *)buf = IPMAXPACKET;
	return sizeof(int);
	
      case GETOPTPACKET:
	/* 
	 * A somewhat meaningless question to be asking the protocol.
	 * It makes more sense to ask an individual session that knows
	 * about the MTU.
	 */
	checkLen(len, sizeof(int));
	*(int *)buf = IPOPTPACKET - IPHLEN;
#ifdef IPSEC
	if (self->numdown>1) {
	  int delta = xControl(xGetDown(self, IPSEC_PROTL), opcode, 0, 0);

	  xTrace0(ipp, TR_SOFT_ERROR, "ip control: over ipsec; GETOPTPACKET on protocol is not well-defined");
	  if (delta>0) *(int *)buf -= delta;
	}
#endif /* IPSEC */
	return sizeof(int);
	
	/* test control ops - remove later */
      case IP_GETRTINFO:
	/* get route info for a given dest address :
	   in : IP host address 
	   out : route structure for this address
	   */
	{
	    xkern_return_t	xkr;
	    
	    checkLen(len, sizeof(route));
	    dest = *(IPhost *)buf;
	    xkr = rt_get(&pstate->rtTbl, &dest, (route *)buf);
	    return (xkr == XK_SUCCESS) ? sizeof(route) : -1;
	}

#ifdef IPSEC
      case IP_PSEUDOHDR:   /* Turn on flag saying to fixup length on IPpseudoheader. */
	if (self->numdown>1)
	  return xControl(xGetDown(self, IPSEC_PROTL), opcode, buf, len);
	else return 0;
#else
      case IP_PSEUDOHDR:
	return 0;
#endif /* IPSEC */

      default:
	xTrace0(ipp,3,"Unrecognized opcode");
	return xControl(xGetDown(self, 0), opcode, buf, len);
    }
}


